import { ErrorMapper } from 'utils/ErrorMapper';
import { runAsHarvester } from './creep/harvester';
import { runAsBuilder } from './creep/builder';
import { runAsUpgrader } from './creep/upgrader';
import { uniqId } from './utils/Identity';

function makeName(): string {
  return uniqId();
}

enum CreepRole {
  HARVESTER = 'harvester',
  UPGRADER = 'upgrader',
  BUILDER = 'builder'
}

function makeCreep(role: CreepRole, count: number) {
  const creeps = Object.values(Game.creeps).filter((c) => c.memory.role === role);
  console.log(`Make creep "${role}"`, 'need', count, 'has', creeps.length);
  if (creeps.length < count) {
    const firstSpawn = _.first(Object.values(Game.spawns));
    const capacity = firstSpawn.room.energyCapacityAvailable;
    const body = selectBody(capacity);
    const name = makeName();
    const memory = { role: role, sourceId: '' } as CreepMemory;
    const err = firstSpawn.spawnCreep(body, name, {
      memory,
      directions: [BOTTOM, BOTTOM_RIGHT, RIGHT, BOTTOM_LEFT, LEFT]
    });
    console.log(`Make creep "${role}"`, 'body', calcBodyCost(body), 'err', err);
  }
}

function selectBody(capacity: number): Array<BodyPartConstant> {
  const bodies = [
    [WORK, WORK, CARRY, CARRY, CARRY, CARRY, MOVE, MOVE, MOVE, MOVE],
    [WORK, WORK, CARRY, CARRY, CARRY, MOVE, MOVE, MOVE],
    [WORK, WORK, CARRY, CARRY, MOVE, MOVE],
    [WORK, CARRY, CARRY, MOVE, MOVE]
  ];
  const fallback = [WORK, CARRY, MOVE];
  for (let body of bodies) {
    const cost = calcBodyCost(body);
    if (cost < capacity) {
      return body;
    }
  }
  return fallback;
}

function calcBodyCost(body: Array<BodyPartConstant>): number {
  return _.sum(body.map((p) => BODYPART_COST[p]));
}

function callHarvestersFromOthers(minHarvCount: number) {
  const harvesters = Object.values(Game.creeps).filter((c) => c.memory.role === CreepRole.HARVESTER);
  const curHarvCount = harvesters.length;
  if (curHarvCount < minHarvCount) {
    const others = Object.values(Game.creeps).filter((c) => c.memory.role !== CreepRole.HARVESTER);
    const required = Math.min(minHarvCount - curHarvCount, others.length);
    console.log('Call harvesters', required);
    let count = 0;
    for (let creep of others) {
      if (count >= required) {
        break;
      }
      count++;
      creep.memory.role = CreepRole.HARVESTER;
    }
  }
}

// When compiling TS to JS and bundling with rollup, the line numbers and file names in error messages change
// This utility uses source maps to get the line numbers and file names of the original, TS source code
export const loop = ErrorMapper.wrapLoop(() => {
  console.log('');
  console.log(`Current game tick is ${Game.time}, ${Game.cpu.bucket}`);

  const room = _.first(Object.values(Game.rooms));
  if (room) {
    console.log('Room', room.name, room.energyCapacityAvailable, room.energyAvailable);
    const roads = room.find(FIND_STRUCTURES, {
      filter: (s) => s.structureType === STRUCTURE_ROAD && s.hits < s.hitsMax
    });
    const minRoadHits = _.min(roads.map((r) => r.hits));
    const structures = room.find(FIND_STRUCTURES, {
      filter: (s) => s.structureType !== STRUCTURE_ROAD && s.hits < s.hitsMax
    });
    const minStructHits = _.min(structures.map((r) => r.hits));
    console.log('Repairs', 'roads', minRoadHits, 'structs', minStructHits);
  }

  // Automatically delete memory of missing creeps
  for (const name in Memory.creeps) {
    if (!(name in Game.creeps)) {
      delete Memory.creeps[name];
    }
  }

  callHarvestersFromOthers(4);

  makeCreep(CreepRole.HARVESTER, 4);
  makeCreep(CreepRole.UPGRADER, 4);
  makeCreep(CreepRole.BUILDER, 4);

  // Process current creeps
  for (let name in Game.creeps) {
    const creep = Game.creeps[name];
    if (creep.memory.role === CreepRole.HARVESTER) {
      runAsHarvester(creep);
    }
    if (creep.memory.role === CreepRole.UPGRADER) {
      runAsUpgrader(creep);
    }
    if (creep.memory.role === CreepRole.BUILDER) {
      runAsBuilder(creep);
    }
  }

  for (let room of Object.values(Game.rooms)) {
    for (let structure of room.find(FIND_MY_STRUCTURES)) {
      if (structure.structureType === STRUCTURE_TOWER) {
        runTower(structure);
      }
    }
  }
});

function runTower(tower: StructureTower) {
  const room = tower.room;
  const hostile = room.find(FIND_HOSTILE_CREEPS);
  if (hostile.length > 0) {
    const res = tower.attack(_.first(hostile));
    console.log('Attack', res);
    return;
  }

  const creeps = room.find(FIND_MY_CREEPS, {
    filter: (c) => c.hits < c.hitsMax
  });
  if (creeps.length > 0) {
    const res = tower.heal(_.first(creeps));
    console.log('Heal', res);
    return;
  }

  // Energy only for attack and heal
  if (tower.store[RESOURCE_ENERGY] < 500) {
    return;
  }

  const roads = room.find(FIND_STRUCTURES, {
    filter: (s) => s.structureType === STRUCTURE_ROAD && s.hits < s.hitsMax
  });
  if (roads.length > 0) {
    roads.sort((r1, r2) => r1.hits - r2.hits);
    let road = _.first(roads);
    const res = tower.repair(road);
    console.log('Repair', res);
    return;
  }

  const structures = room.find(FIND_STRUCTURES, {
    filter: (s) => s.hits < s.hitsMax
  });
  if (structures.length > 0) {
    structures.sort((r1, r2) => r1.hits - r2.hits);
    let structure = _.first(structures);
    const res = tower.repair(structure);
    console.log('Repair', res);
    return;
  }
}
